use super::*;
use crate::parser::Token;
use crate::Node;

fn zero() -> GaussianRational {
    GaussianRational::new(Rational::new(0, 1), Rational::new(0, 1))
}

fn one() -> GaussianRational {
    GaussianRational::new(Rational::new(1, 1), Rational::new(0, 1))
}

fn minus_one() -> GaussianRational {
    GaussianRational::new(Rational::new(-1, 1), Rational::new(0, 1))
}

fn add(lhs: Vec<GaussianRational>, rhs: Vec<GaussianRational>) -> Vec<GaussianRational> {
    let mut res = Vec::new();
    let mut len = lhs.len();
    if rhs.len() > len {
        len = rhs.len();
    }
    for _ in 0..len {
        res.push(zero());
    }

    for (i, gr) in lhs.iter().enumerate() {
        res[i] = res[i] + *gr;
    }
    for (i, gr) in rhs.iter().enumerate() {
        res[i] = res[i] + *gr;
    }

    res
}

fn sub(lhs: Vec<GaussianRational>, rhs: Vec<GaussianRational>) -> Vec<GaussianRational> {
    let mut res = Vec::new();
    let mut len = lhs.len();
    if rhs.len() > len {
        len = rhs.len();
    }
    for _ in 0..len {
        res.push(zero());
    }

    for (i, gr) in lhs.iter().enumerate() {
        res[i] = res[i] + *gr;
    }
    for (i, gr) in rhs.iter().enumerate() {
        res[i] = res[i] - *gr;
    }

    res
}

fn mul(lhs: Vec<GaussianRational>, rhs: &Vec<GaussianRational>) -> Vec<GaussianRational> {
    println!("\nIci ca multiplie {:?} et {:?}", lhs, rhs);
    let len = lhs.len() + rhs.len() - 1;
    let mut res = Vec::new();
    for _ in 0..len {
        res.push(zero());
    }
    for (i, lhs) in lhs.iter().enumerate() {
        for (j, rhs) in rhs.iter().enumerate() {
            res[i + j] = res[i + j] + *lhs * *rhs;
        }
    }
    res
}

fn exp(lhs: Vec<GaussianRational>, rhs: Vec<GaussianRational>) -> Result<Vec<GaussianRational>, String> {
    if rhs.len() != 1 {
        return Err(format!("exponential only supports natural exponents, unknown exponents aren't supported"));
    }
    if rhs[0].imaginary.numerator != 0 {
        return Err(format!("exponential only supports natural exponents, complex exponents aren't supported"));
    }
    if rhs[0].real.denominator != 1 {
        return Err(format!("exponential only supports natural exponents, rational exponents aren't supported"));
    }
    let mut res = Vec::new();
    let pow = rhs[0].real.numerator;
    res.push(one());
    for _ in 0..pow {
        res = mul(res, &lhs);
    }
    Ok(res)
}

fn check_div(lhs: &Vec<GaussianRational>, rhs: &Vec<GaussianRational>) -> bool {
    for i in 0..(rhs.len() - 1) {
        if rhs[i] != zero() || lhs[i] != zero() {
            return false
        }
    }
    true
}

fn div(lhs: Vec<GaussianRational>, rhs: Vec<GaussianRational>) -> Result<Vec<GaussianRational>, String> {
    if rhs.len() > lhs.len() {
        return Err(format!("division supported if the divisor to be of same or smaller degree than the dividend."));
    }
    if !check_div(&lhs, &rhs) {
        return Err(format!("division supported if the divisor has only one non zero coefficient"));
    }

    let denominator = *rhs.last().unwrap();
    let len = lhs.len() - rhs.len() + 1;
    let mut res = Vec::new();
    for _ in 0..len {
        res.push(zero());
    }

    for i in (rhs.len() - 1)..lhs.len() {
        let res_i = i - (rhs.len() - 1);
        res[res_i] = lhs[i] / denominator;
    }

    Ok(res)
}

fn calculate(operator: Token, lhs: Vec<GaussianRational>, rhs: Vec<GaussianRational>) -> Result<Vec<GaussianRational>, String> {
    match operator {
        Token::Addition() => Ok(add(lhs, rhs)),
        Token::Substraction() => Ok(sub(lhs, rhs)),
        Token::Multiplication() => Ok(mul(lhs, &rhs)),
        Token::Division() => div(lhs, rhs),
        Token::Exponentiation() => exp(lhs, rhs),
        Token::Equal() => Ok(sub(lhs, rhs)),
        Token::Modulo() => Err(format!("modulo isn't suported yet by this program.")),
        _ => unreachable!(),
    }
}

fn calculate_unary(operator: Token, operand: Vec<GaussianRational>) -> Vec<GaussianRational> {
    match operator {
        Token::Addition() => operand,
        Token::Substraction() => mul(vec![minus_one()], &operand),
        _ => unreachable!(),
    }
}

pub fn evaluate(ast: Node) -> Result<Vec<GaussianRational>, String> {
    match ast {
        Node::Binary { operator, lhs, rhs } => calculate(operator, evaluate(*lhs)?, evaluate(*rhs)?),
        Node::Unary { operator, operand } => Ok(calculate_unary(operator, evaluate(*operand)?)),
        Node::Leaf(value) => Ok(value),
    }
}

pub fn is_equation(ast: &Node) -> bool {
    match ast {
        Node::Binary { operator, lhs: _, rhs: _ } => *operator == Token::Equal(),
        _ => false,
    }
}
